package Mentat::Message::Correlator::Engine;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Correlator::Engine - Short module description

=head1 SYNOPSIS

 use Mentat::Message::Correlator::Engine;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Mentat::Core qw(:c_general);
use Config::Parser;
use Data::Filter::Factory;
use Data::Filter::Compiler;
use Data::Query::Object;

use Mentat::Message::Value::IDEA;
use Mentat::Message::Uniter;
use Mentat::Message::Correlator::Context;
use Mentat::Message::Correlator::Context::Queue::Memory;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Directive processor callback for parsing config file
sub directive_processor
{
    my ($section_name, $section_id, $directive, $value) = @_;
    my ($error, $errmsg) = (0, undef);

    # Prepare 'rule' value
    if ($directive eq 'rule') {
        $value = [split(/:/, $value)];
    }
    return ($directive, $value, $error, $errmsg);
}

# Parameters for rule parser
my $RULE_PARSER_PARAMS = {
        'cb_directive_processor' => \&directive_processor,
        'stack_values'           => "key rule",
        'compress'               => 2,
        'unique_sections'        => 1,
        'structure'              => {__general__    => {},
                                     ruleset        => {filter      => 'string',
                                                        key         => 'string',
                                                        description => 'string',
                                                        mode        => 'string',
                                                        thresh_min  => 'int',
                                                        thresh_max  => 'int',
                                                        window      => 'int',
                                                        ttl         => 'int',
                                                        compile     => 'string',
                                                        rule        => 'string',
                                               },
                                },
            };

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Message::Correlator::Engine->new();
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Message::Correlator::Engine reference on success, croaks on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    ### Creating Mentat.Message.Correlator.Engine instance...
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat.Message.Correlator.Engine instance...
}

=item correlate($) [PUBLIC]

  Usage     : my $message = $instance->correlate($message);
  Purpose   : Correlate given message
  Arguments : Mentat::Message $message - message to be correlated [MANDATORY]
  Returns   : ARRAY containing correlated Mentat::Messages
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub correlate($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message,) = @_;

    my @result;
    my ($rs, $r, $context_key, $context, $cmessage, $result);

    # Check all correlation rules in sequence
    foreach my $rsn (keys %{$self->{RULES}}) {
        $rs = $self->{RULES}->{$rsn};

        # Message must match correlation rule filter
        if ($self->_check_filters($rs, $message)) {
            print STDERR "Message passed filters for ruleset '$rsn'\n" if $DEVEL;

            # Fetch the appropriate correlation context and update it
            $context = $self->fetch_context($rs, $message);
            next unless $context;
            $cmessage = $context->message;
            if ($cmessage) {
                print STDERR "Uniting\n" if $DEVEL;
                $message = $self->{UNITER}->unite_into($rs->{rule}, $cmessage, $message);
            }
            $r = $context->update($message);

            # If the context condition matched, release the correlated message
            if ($r) {
                print STDERR "Context condition matched\n" if $DEVEL;
                push(@result, $context->message);
            }
            else {
                print STDERR "Storing context\n" if $DEVEL;
                $self->store_context($context);
            }
        }
        else {
            print STDERR "Message was filtered out from ruleset '$rsn'\n" if $DEVEL;
        }
    }
    return @result;
}

=item check_contexts() [PUBLIC]

  Usage     : my $message = $instance->check_contexts();
  Purpose   : Check for timeouted contexts
  Arguments : None
  Returns   : Mentat::Message $message if any context timeouted, UNDEF if there is not any
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub check_contexts()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $rv;
    my $context;
    while ($context = $self->{Q_CONTEXTS}->dequeue()) {
        print STDERR "CD: $context\n" if $DEVEL;
        next unless $context->condition_match(TRUE);
        $rv = $context->message;
        last;
    }
    return $rv;
}

#-------------------------------------------------------------------------------

=item fetch_context($$) [PUBLIC]

  Usage     : my $context = $instance->fetch_context($rs, $message);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub fetch_context($$)
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
    my ($ruleset, $alert_message) = @_;

    # Calculate the appropriate context key
    my $context_key = $self->_get_context_key($ruleset, $alert_message);
    return undef unless $context_key;

    my $context = $self->{Q_CONTEXTS}->fetch($context_key);
    unless ($context) {
        $context = Mentat::Message::Correlator::Context->new(id => $context_key, %$ruleset);
    }
    return $context;
}

=item store_context($) [PUBLIC]

  Usage     : $instance->store_context($context);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub store_context($)
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
    my ($context,) = @_;

    return $self->{Q_CONTEXTS}->enqueue($context);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Message::Correlator::Engine instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Message::Correlator::Engine $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat.Message.Correlator.Engine instance...
    my ($filter_grammar_file, $correlator_config) = @_;

    # Context filter related stuff
    $self->{QUERY}     = Data::Query::Object->new(dt_resolver => Mentat::Message::Value::IDEA->new());
    $self->{FFACTORY}  = Data::Filter::Factory->new_from_file($filter_grammar_file);
    $self->{FCOMPILER} = Data::Filter::Compiler->new();
    $self->{COMPILE}   = 'idea';

    # Load correlator configurations from external file
    $self->{RULES} = $self->_load_correlator_config($correlator_config);

    foreach my $r (keys %{$self->{RULES}}) {
        $self->{RULES}->{$r}->{filter} = $self->_build_filters($self->{RULES}->{$r}->{filter});
        die "Correlation ruleset '$r' is missing context key\n" unless $self->{RULES}->{$r}->{key};
        die "Correlation ruleset '$r' is missing context TTL\n" unless $self->{RULES}->{$r}->{ttl};
        $self->{RULES}->{$r}->{ttl}        = (defined $self->{RULES}->{$r}->{ttl})?int($self->{RULES}->{$r}->{ttl}):undef;
        $self->{RULES}->{$r}->{window}     = (defined $self->{RULES}->{$r}->{window})?int($self->{RULES}->{$r}->{window}):undef;
        $self->{RULES}->{$r}->{thresh_min} = (defined $self->{RULES}->{$r}->{thresh_min})?int($self->{RULES}->{$r}->{thresh_min}):0;
        $self->{RULES}->{$r}->{thresh_max} = (defined $self->{RULES}->{$r}->{thresh_max})?int($self->{RULES}->{$r}->{thresh_max}):undef;
    }

    # Context storage queue
    $self->{Q_CONTEXTS} = Mentat::Message::Correlator::Context::Queue::Memory->new();

    # Message uniter object
    $self->{UNITER} = Mentat::Message::Uniter->new();

    return $self;
}

# _load_correlator_config($) [PROTECTED]
#
#  Usage     : my $config = $self->_load_correlator_config($corelator_config);
#  Purpose   : Load and prepare filter configurations
#  Arguments : STRING or HASH REFERENCE $correlator_config - file name or bare configurations
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _load_correlator_config($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($correlator_config,) = @_;

    my $config = $correlator_config;

    # If the configurations were not given bare
    unless (ref $correlator_config eq 'HASH') {
        my $parser = Config::Parser->new($RULE_PARSER_PARAMS);
        $config    = $parser->config($correlator_config)->{ruleset};

        die "Rule file contains errors: " . join(',', @{$parser->get_errors()}) . "\n" unless $config;
    }

    return $config;
}

# _build_filters($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _build_filters($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter_rules,) = @_;

    my $f;
    my $filters = [];
    foreach my $rule (@{$filter_rules}) {
        $f = $self->{FFACTORY}->filter_from_string($rule);
        $f = $self->{FCOMPILER}->compile($f, $self->{COMPILE}) if $self->{COMPILE};
        push(@$filters, $f);
    }
    return $filters;
}

# _check_filters($$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Filter the given message through selected filter chain
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _check_filters($$)
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
    my ($ruleset, $alert_message) = @_;

    # Walk through all filters in selected chain
    foreach my $filter (@{$ruleset->{filter}})  {
        print STDERR "Checking filter '$filter'\n" if $DEVEL; #-+-> DEVEL ONLY <-+-#

        # Filter the message
        my $result = $self->{QUERY}->is_satisfied_by($filter, $alert_message);
        unless ($result) {
            # Message did not matched some filter in the chain, reject
            return 0;
        }
    }
    # Message matched all filters in chain, accept
    return 1;
}

# _get_context_key($$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Filter the given message through selected filter chain
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _get_context_key($$)
{
    my $self = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;
    my ($ruleset, $alert_message) = @_;

    my @context_key = ();
    my $values;
    foreach my $key (@{$ruleset->{key}})    {
        $values = $alert_message->path_values($key);
        return undef unless @$values;
        push(@context_key, @$values)
    }
    return join(':', @context_key);
}

=pod

=back

=cut

1;
