package Mentat::Stats::Charter;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Stats::Charter - Short module description

=head1 SYNOPSIS

 use Mentat::Stats::Charter;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Stats(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Storable 'dclone';

use Number::Format qw(:subs :vars);

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use GD::Graph::Data;   #libgd-graph-perl
use GD::Graph::pie;
use GD::Graph::bars;
use GD::Graph::hbars;
use GD::Graph::lines;
use GD::Graph::area;

#-- Custom application modules ------------------------------------------------#
use Color::Palette;
use Value::Convertor;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant CHART_TTL => 1;

use constant CHART_WIDTH  => 600;
use constant CHART_HEIGHT => 400;

use constant CHART_SL_WIDTH  => 150;
use constant CHART_SL_HEIGHT => 100;

use constant CHART_SLS_WIDTH  => 90;
use constant CHART_SLS_HEIGHT => 60;

#use constant DFLT_COLOUR => '#46494C';
use constant DFLT_COLOUR => '#003F7F';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Definition of chart sizes
my %CHART_SIZES = (
    'hbar' => [CHART_WIDTH(),  CHART_HEIGHT()],
    'bar'  => [CHART_WIDTH(),  CHART_HEIGHT()],
    'pie'  => [CHART_HEIGHT(), CHART_HEIGHT()],
    'sl_hbar' => [CHART_SL_WIDTH(),  CHART_SL_HEIGHT()],
    'sl_bar'  => [CHART_SL_WIDTH(),  CHART_SL_HEIGHT()],
    'sl_pie'  => [CHART_SL_HEIGHT(), CHART_SL_HEIGHT()],
    'sls_hbar' => [CHART_SLS_WIDTH(),  CHART_SLS_HEIGHT()],
    'sls_bar'  => [CHART_SLS_WIDTH(),  CHART_SLS_HEIGHT()],
    'sls_pie'  => [CHART_SLS_HEIGHT(), CHART_SLS_HEIGHT()],
);

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC ,STATIC]

  Usage     : my $instance = Mentat::Stats::Charter->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Stats::Charter reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat::Stats::Charter instance...
}

#-------------------------------------------------------------------------------

=item chart_hbar($$$;$)     [PUBLIC]
=item chart_hbar_sl($$$;$)  [PUBLIC]
=item chart_hbar_sls($$$;$) [PUBLIC]
=item chart_bar($$$;$)      [PUBLIC]
=item chart_bar_sl($$$;$)   [PUBLIC]
=item chart_bar_sls($$$;$)  [PUBLIC]
=item chart_pie($$$;$)      [PUBLIC]
=item chart_pie_sl($$$;$)   [PUBLIC]
=item chart_pie_sls($$$;$)  [PUBLIC]

  Usage     : $chart = $charter->chart_hbar($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_hbar_sl($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_hbar_sls($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_bar($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_bar_sl($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_bar_sls($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_pie($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_pie_sl($chart, $file_id, $conv, $replace_names);
              $chart = $charter->chart_pie_sls($chart, $file_id, $conv, $replace_names);
  Purpose   : Generate chart from given data
  Arguments : HASH REFERENCE $chart [MANDATORY]
              STRING $file_id [MANDATORY]
              STRING $conv [MANDATORY]
              BOOL $replace_names [OPTIONAL]
  Returns   : HASH REFERENCE $chart
  Throws    : Dies, if not invoked on instance

=cut

sub chart_hbar($$$;$)     { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_hbar_sl($$$;$)  { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_hbar_sls($$$;$) { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_bar($$$;$)      { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_bar_sl($$$;$)   { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_bar_sls($$$;$)  { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_pie($$$;$)      { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_pie_sl($$$;$)   { my $self = shift; return $self->_chart_gd('hbar', @_); }
sub chart_pie_sls($$$;$)  { my $self = shift; return $self->_chart_gd('hbar', @_); }

=item chart($$$$;$)     [PUBLIC]

  Usage     : $chart = $charter->chart($type, $chart, $file_id, $conv, $replace_names);
  Purpose   : Generate chart from given data
  Arguments : STRING $type [MANDATORY]
              HASH REFERENCE $chart [MANDATORY]
              STRING $file_id [MANDATORY]
              STRING $conv [MANDATORY]
              BOOL $replace_names [OPTIONAL]
  Returns   : HASH REFERENCE $chart
  Throws    : Dies, if not invoked on instance

=cut

sub chart($$$$;$) { my $self = shift; return $self->_chart_gd(@_); }

#-------------------------------------------------------------------------------

=item colpal($) [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub colpal($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($size,) = @_;

    my @palette = ();
    my @tmp = Color::Palette->palette($size);

    foreach my $c (@tmp) {
        my ($color, $rgb) = @$c;
        push(@palette, "#"."$rgb");
    }
    Log::Loger->info(__PACKAGE__, "[Stats::Charter] Palette: @palette") if $DEVEL;
    return \@palette;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Stats::Charter instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Stats::Charter $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat::Stats::Charter instance...
    my %config = @_;

    die "[Stats::Charter] Chart directory '$config{'charts_dir'}' does not exist\n" unless -d $config{charts_dir};
    die "[Stats::Charter] Chart directory '$config{'charts_dir'}' is not writable\n" unless -w $config{charts_dir};

    $self->{CHARTS_DIR} = $config{charts_dir};
    $self->{TTL}        = ($config{ttl}) ? int($config{ttl}) : CHART_TTL;

    if ($self->{CHARTS_DIR} =~ /(.*)/) { $self->{CHARTS_DIR} = $1; }

    return $self;
}

# _save_image($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _save_image($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filename, $data) = @_;

    open(IMG, ">$filename") or die $!;
    binmode IMG;
    print IMG $data;
    close(IMG);

    Log::Loger->info(__PACKAGE__, "[Stats::Charter] Stored chart into file '$filename'");

    return $filename;
}

#-------------------------------------------------------------------------------

sub _c_fn($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return format_number(@_);
}

sub _c_dur($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return Value::Convertor->duration_to_string(@_);
}

#-------------------------------------------------------------------------------
# GD::GRAPH charting
#-------------------------------------------------------------------------------

sub _chart_gd_data_preprocess($$$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $chart_data, $conv, $replace_labels) = @_;

    # No preprocessing needed for pie charts
    return $chart_data if $type =~ /pie|sl_pie|sls_pie/;

    my $values = dclone($chart_data);
    for (my $i = 0; $i < @{$values->[1]}; $i++) {
        # Optionally replace labels with sequential number
        if ($replace_labels) {
            $values->[0]->[$i]     = ($i + 1);
            $chart_data->[0]->[$i] = ($i + 1);
        }
        # Convert the displayed values to desired format
        $values->[1]->[$i] = $self->$conv($values->[1]->[$i]);
    }
    return $values;
}

sub _chart_gd_init($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $chart_data) = @_;
    confess "Unknown chart type '$type', unable to determine the chart size" unless exists $CHART_SIZES{$type};

    # Create the chart manipulation object
    my $chart_object;
    if    ($type =~ /hbar|sl_hbar|sls_hbar/) {
        $chart_object = GD::Graph::hbars->new(@{$CHART_SIZES{$type}});
    }
    elsif ($type =~ /bar|sl_bar|sls_bar/) {
        $chart_object = GD::Graph::bars->new(@{$CHART_SIZES{$type}});
    }
    elsif ($type =~ /pie|sl_pie|sls_pie/) {
        $chart_object = GD::Graph::pie->new(@{$CHART_SIZES{$type}});
    }
    confess "Unknown chart type '$type', unable to initialize chart object" unless $chart_object;

    my $dsize = @{$chart_data->[1]};
    # Hide axis and everything not necessary for sparkline charts
    if ($type =~ /sl_hbar|sls_hbar|sl_bar|sls_bar/) {
        $chart_object->set(
            fgclr        => undef,
            labelclr     => undef,
            axislabelclr => undef,
            legendclr    => undef,
            valuesclr    => undef,
            textclr      => undef,
            accentclr    => undef,
            no_axes      => 1,
            axis_space   => 1,
            text_space   => 1,
            dclrs        => $self->colpal($dsize),
        );
    }
    # Setup the colours for all other charts
    else {
        $chart_object->set(
            fgclr        => DFLT_COLOUR(),
            labelclr     => DFLT_COLOUR(),
            axislabelclr => DFLT_COLOUR(),
            legendclr    => DFLT_COLOUR(),
            valuesclr    => DFLT_COLOUR(),
            textclr      => DFLT_COLOUR(),
            accentclr    => DFLT_COLOUR(),
            #accentclr    => "#FFFFFE",
            dclrs        => $self->colpal($dsize),
        );
    }

    # Additional setup for pie charts
    if ($type =~ /pie|sl_pie|sls_pie/) {
        $chart_object->set(
            #accentclr      => DFLT_COLOUR(),
            suppress_angle => 360,
            '3d'           => 0,
        );
    }
    # Additional setup for bar charts
    elsif ($type =~ /hbar|bar/) {
        $chart_object->set(
            show_values => $chart_data,
        );
    }

    return $chart_object;
}

sub _chart_gd($$$$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $chart, $file_id, $conv, $replace_names) = @_;

    unless ($chart) {
        Log::Loger->info(__PACKAGE__, "[Stats::Charter] Unable to generate GD '$type' chart '$file_id', no data");
        return $chart;
    }

    my $suffix     = ($type =~ /^(sls?)_/) ? '.'.$1 : '';
    my $chart_data = $chart->{data};
    my $file_name  = $file_id.$suffix.".png";
    my $file_path  = join('/', $self->{CHARTS_DIR}, $file_name);
    my $ts         = (-f $file_path)?((stat($file_path))[9]):0;
    my $ttl        = time - $self->{TTL};
    my $conversion = "_c_$conv";

    if (not -f $file_path or ($ts < $ttl)) {
        # Preprocess the chart data values (format conversions, label replacements)
        my $values = $self->_chart_gd_data_preprocess($type, $chart_data, $conversion, $replace_names);
        # Initialize the chart object
        my $chart_object = $self->_chart_gd_init($type, $values);

        # Plot the chart data
        Log::Loger->debug(__PACKAGE__, "[Stats::Charter] Plotting chart data: ".Dumper($chart_data)) if $DEVEL;
        my $chart_image = $chart_object->plot($chart_data);

        # Save the chart into external file
        if ($chart_image) {
            $self->_save_image($file_path, $chart_image->png());
            Log::Loger->info(__PACKAGE__, "[Stats::Charter] Regenerated GD '$type' chart '$file_name'");
        }
        else {
            Log::Loger->error(__PACKAGE__, "[Stats::Charter] Unable to generate GD '$type' chart '$file_name' with data: ".Dumper($chart_data));
        }
    }
    else {
        Log::Loger->info(__PACKAGE__, "[Stats::Charter] GD '$type' chart '$file_name' already exists and is current enough");
    }
    $chart->{'ts'}           = time;
    $chart->{'type'}         = $type;
    $chart->{'chart_'.$type} = $file_name;

    $chart->{'file'}     = $file_name if not $type =~ /^sl/;
    $chart->{'file_sl'}  = $file_name if $type =~ /^sl_/;
    $chart->{'file_sls'} = $file_name if $type =~ /^sls_/;

    return $chart;
}

=pod

=back

=cut

1;
